# This file is part of CMFNotification
#
# Copyright (c) 2005-2006 by Pilot Systems (http://www.pilotsystems.net)
# 
# CMFNotification is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#
"""Define notification tool.

$Id: NotificationTool.py 49219 2007-09-14 07:15:34Z dbaty $
"""

import re
import inspect
import logging
from email.Header import Header
from types import StringType

from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from ZODB.POSException import ConflictError, POSKeyError
from OFS.PropertyManager import PropertyManager
try:
    from persistent.mapping import PersistentMapping
except ImportError:
    ## Probably Zope 2.7.
    from Persistence import PersistentMapping
from AccessControl.PermissionRole import rolesForPermissionOn
from AccessControl import ClassSecurityInfo, getSecurityManager, Unauthorized

from Products.CMFCore.permissions import ManagePortal
from Products.CMFCore.utils import UniqueObject, getToolByName
from Products.CMFCore.Expression import Expression

from utils import getExpressionContext
from permissions import SUBSCRIBE_PERMISSION
from exceptions import MailHostNotFound, DisabledFeature, InvalidEmailAddress

ID = 'portal_notification'
TITLE = 'CMF Notification tool'
META_TYPE = 'CMFNotificationTool'

## This regexp was taken from Silva ('Silva/subscriptionservice.py').
EMAIL_REGEXP = re.compile('^[0-9a-zA-Z_&.%+-]+@([0-9a-zA-Z]([0-9a-zA-Z-]*[0-9a-zA-Z])?\.)+[a-zA-Z]{2,6}$')
RULE_DELIMITER = '::'
LOG = logging.getLogger('CMFNotification')
MAIL_HOST_META_TYPES = ('Mail Host', 'Secure Mail Host', 'Maildrop Host')
## Ignore objects temporarily created by the Plone Factory Tool
DEFAULT_IGNORE_RULES = ["python: getattr(context, 'isTemporary', lambda: False)()"]


class NotificationTool(UniqueObject, SimpleItem, PropertyManager):
    """Main notification tool."""
    id = ID
    title = TITLE
    meta_type = META_TYPE

    manage_options = (PropertyManager.manage_options
                      + SimpleItem.manage_options)

    ## Extra subscriptions
    extra_subscriptions_enabled = False
    ## FIXME: rename this to something like
    ## 'restrict_subscriptions_to_authenticated_users'
    extra_subscriptions_for_authenticated_only = True
    extra_subscriptions_recursive = True

    ## Debug settings
    debug_log_addresses = False

    ## Ignore rules
    ignore_rules = DEFAULT_IGNORE_RULES

    ## Item creation
    item_creation_notification_enabled = True
    on_item_creation_users = []
    on_item_creation_mail_template = []
    ## Item modification
    item_modification_notification_enabled = True
    on_item_modification_users = []
    on_item_modification_mail_template = []
    ## Workflow transition
    wf_transition_notification_enabled = True
    on_wf_transition_users = []
    on_wf_transition_mail_template = []
    ## Member registration
    member_registration_notification_enabled = True
    on_member_registration_users = []
    on_member_registration_mail_template = []
    ## Member modification
    member_modification_notification_enabled = True
    on_member_modification_users = []
    on_member_modification_mail_template = []
    ## Discussion item creation
    discussion_item_creation_notification_enabled = True
    on_discussion_item_creation_users = []
    on_discussion_item_creation_mail_template = []

    _properties = ({'id': 'extra_subscriptions_enabled',
                    'label': 'Enable extra subscriptions',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'extra_subscriptions_for_authenticated_only',
                    'label': 'Enable extra subscriptions only for authenticated users',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'extra_subscriptions_recursive',
                    'label': 'Toggle recursive mode for extra subscriptions',
                    'mode': 'w',
                    'type': 'boolean'},

                   {'id': 'debug_log_addresses',
                    'label': 'Toggle debug mode: log addresses',
                    'mode': 'w',
                    'type': 'boolean'},

                   {'id': 'ignore_rules',
                    'label': 'Rules (ignore)',
                    'mode': 'w',
                    'type': 'lines'},

                   {'id': 'item_creation_notification_enabled',
                    'label': 'Enable item creation notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_item_creation_users',
                    'label': 'Rules on item creation (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_item_creation_mail_template',
                    'label': 'Rules on item creation (mail template)',
                    'mode': 'w',
                    'type': 'lines'},
 
                   {'id': 'item_modification_notification_enabled',
                    'label': 'Enable item modification notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_item_modification_users',
                    'label': 'Rules on item modification (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_item_modification_mail_template',
                    'label': 'Rules on item modification (mail template)',
                    'mode': 'w',
                    'type': 'lines'},

                   {'id': 'wf_transition_notification_enabled',
                    'label': 'Enable workflow transition notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_wf_transition_users',
                    'label': 'Rules on workflow transition (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_wf_transition_mail_template',
                    'label': 'Rules on workflow transition (mail template)',
                    'mode': 'w',
                    'type': 'lines'},

                   {'id': 'member_registration_notification_enabled',
                    'label': 'Enable member registration notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_member_registration_users',
                    'label': 'Rules on member registration (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_member_registration_mail_template',
                    'label': 'Rules on member registration (mail template)',
                    'mode': 'w',
                    'type': 'lines'},

                   {'id': 'member_modification_notification_enabled',
                    'label': 'Enable member modification notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_member_modification_users',
                    'label': 'Rules on member modification (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_member_modification_mail_template',
                    'label': 'Rules on member modification (mail template)',
                    'mode': 'w',
                    'type': 'lines'},

                   {'id': 'discussion_item_creation_notification_enabled',
                    'label': 'Enable discussion item creation notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_discussion_item_creation_users',
                    'label': 'Rules on discussion item creation (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_discussion_item_creation_mail_template',
                    'label': 'Rules on discussion item creation (mail template)',
                    'mode': 'w',
                    'type': 'lines'},
                   )

    security = ClassSecurityInfo()
    decPrivate = security.declarePrivate
    decProtected = security.declareProtected
    decPublic = security.declarePublic


    def __init__(self, *args, **kwargs):
        self._uid_to_path = PersistentMapping()
        self._subscriptions = PersistentMapping()


    #################################################################
    ## Notification handlers
    ########################
    decPrivate('onItemCreation')
    def onItemCreation(self, obj):
        """Handler called when an item is created.

        It returns the number of mails which have been sent.

        **Warning:** this handler is not called when a discussion item
        is added. In this case, ``onDiscussionItemCreation()`` is
        called instead.
        """
        if not self.getProperty('item_creation_notification_enabled'):
            return 0
        if self.ignoreNotification(obj):
            return 0

        extra_bindings = {'current': obj,
                          'previous': None}
        return self._handlerHelper(obj, 'item_creation',
                                   extra_bindings,
                                   extra_bindings,
                                   extra_bindings)


    decPrivate('onItemModification')
    def onItemModification(self, obj):
        """Handler called when an item is modified.

        It returns the number of mails which have been sent.
        """
        if not self.getProperty('item_modification_notification_enabled'):
            return 0
        if self.ignoreNotification(obj):
            return 0

        extra_bindings = {'current': obj,
                          'previous': self.getPreviousVersion(obj)}
        return self._handlerHelper(obj, 'item_modification',
                                   extra_bindings,
                                   extra_bindings,
                                   extra_bindings)


    decPrivate('onWorkflowTransition')
    def onWorkflowTransition(self, workflow, obj, action, result):
        """Handler called when a workflow transition is triggered.

        It returns the number of mails which have been sent.
        """
        if not self.getProperty('wf_transition_notification_enabled'):
            return 0
        if self.ignoreNotification(obj):
            return 0

        wtool = getToolByName(self, 'portal_workflow')
        current_state = wtool.getInfoFor(obj, 'review_state')
        comments = wtool.getInfoFor(obj, 'comments')
        extra_bindings = {'transition': action,
                          'current_state': current_state,
                          'comments': comments}
        return self._handlerHelper(obj, 'wf_transition',
                                   extra_bindings,
                                   extra_bindings,
                                   extra_bindings)


    decPrivate('onMemberRegistration')
    def onMemberRegistration(self, member, properties):
        """Handler called when a new portal member has been registered.

        It returns the number of mails which have been sent.
        """
        if not self.getProperty('member_registration_notification_enabled'):
            return 0
        if self.ignoreNotification(member):
            return 0

        if properties is None:
            properties = {} ## FIXME: How could it be? (Damien)

        current_user = getSecurityManager().getUser()
        extra_bindings = {'current_user': current_user,
                          'member': member,
                          'properties': properties,
                          'event': 'registration'}
        return self._handlerHelper(member, 'member_registration',
                                   extra_bindings,
                                   extra_bindings,
                                   extra_bindings)


    decPrivate('onMemberModification')
    def onMemberModification(self, member):
        """Handler called when a member changes his/her properties.

        It returns the number of mails which have been sent.
        """
        ## This method can also be called when the member is
        ## registered. We have to check that.
        stack = inspect.stack()
        ## 1st item is ourself
        ## 2nd item is 'CMFCore.MemberDataTool.notifyMemberModified()'
        ## 3rd item is 'CMFCore.MemberDataTool.setMemberProperties()'
        ## 4th item is what we want to check: it is either 'addMember'
        ## or 'setProperties()'
        caller = stack[3][3]
        if caller != 'setProperties':
            return 0

        if not self.getProperty('member_modification_notification_enabled'):
            return 0
        if self.ignoreNotification(member):
            return 0

        ## FIXME: what is the purpose of the following lines? (Damien)
        memberdata = getToolByName(self, 'portal_memberdata')
        properties = {}
        for key, value in memberdata.propertyItems():
            properties[key] = value

        current_user = getSecurityManager().getUser()
        extra_bindings = {'current_user': current_user,
                          'member': member,
                          'properties': properties,
                          'event': 'modification'}
        return self._handlerHelper(member, 'member_modification',
                                   extra_bindings,
                                   extra_bindings,
                                   extra_bindings)


    decPrivate('onDiscussionItemCreation')
    def onDiscussionItemCreation(self, discussion_item):
        """Handler called when a discussion item is created.

        It returns the number of mails which have been sent.
        """
        if not self.getProperty('discussion_item_creation_notification_enabled'):
            return 0
        if self.ignoreNotification(discussion_item):
            return 0

        ## We add two bindings to disambiguate the meaning of 'here'
        ## in the mail template and the rules: 'discussion_item' and
        ## 'discussed_item'.
        discussed_item = discussion_item
        while discussed_item.meta_type == discussion_item.meta_type:
            discussed_item = discussed_item.aq_inner.aq_parent.aq_parent
        extra_bindings = {'discussion_item': discussion_item,
                          'discussed_item': discussed_item}
        return self._handlerHelper(discussion_item,
                                   'discussion_item_creation',
                                   extra_bindings,
                                   extra_bindings,
                                   extra_bindings)


    def _handlerHelper(self, obj, what,
                       get_users_extra_bindings,
                       mail_template_extra_bindings,
                       mail_template_options):
        """An helper method for ``on*()`` handlers.

        It returns the number of mails which have been sent.
        """
        self._updateSubscriptionMapping(obj)
        users = self.getUsersToNotify(obj, what,
                                      get_users_extra_bindings)
        if self.isExtraSubscriptionsEnabled():
            users.extend(self.getExtraSubscribersOf(self._getPath(obj)))
        users = self.removeUnAuthorizedSubscribers(users, obj)

        addresses = self.getEmailAddresses(users)
        if not addresses:
            LOG.warning("No addresses for '%s' notification of %s",
                        what, obj.absolute_url(1))
            return 0

        template = self.getMailTemplate(obj, what,
                                        mail_template_extra_bindings)
        if template is None:
            LOG.warning("No mail template for '%s' notification of %s",
                        what, obj.absolute_url(1))
            return 0

        try:
            message = template(**mail_template_options)
        except ConflictError:
            raise
        except:
            LOG.error("Cannot evaluate mail template '%s' on '%s' "\
                      "for '%s'",
                      template.absolute_url(1),
                      obj.absolute_url(1), what,
                      exc_info=True)
            return 0

        return self.sendNotification(addresses, message)
    #################################################################


    #################################################################
    ## Utility methods
    ###############################
    decPrivate('ignoreNotification')
    def ignoreNotification(self, obj):
        """Return ``True`` iff notification have been set to be ignored for
        ``obj``.
        """
        ec = getExpressionContext(obj)
        users = []
        for match_expr in self.getProperty('ignore_rules', ()):
            try:
                if self._match(match_expr, ec):
                    return True
            except ConflictError:
                raise
            except:
                LOG.error("Error in 'ignore_rules' rule "\
                          "('%s') for '%s'",
                          match_expr, obj.absolute_url(1),
                          exc_info=True)
        return False


    decPrivate('getUsersToNotify')
    def getUsersToNotify(self, obj, what, ec_bindings=None):
        """Return a list of users to notify for the ``what`` of an object
        ``obj``, ``what`` being one of the implemented notification
        ("*item_modification*", "*wf_transition*", etc.).

        ``ec_bindings`` is a mapping which is injected into the
        expression context of the expression of the rules.
        """
        rules = self.getProperty('on_%s_users' % what, None)
        if rules is None:
            raise NotImplementedError, \
                'Notification on "%s" is not implemented.' % what

        ec = getExpressionContext(obj, ec_bindings)
        users = []
        for rule in rules:
            try:
                match_expr, users_expr = rule.split(RULE_DELIMITER)
                match_expr, users_expr = match_expr.strip(), users_expr.strip()
            except ValueError:
                LOG.error("'%s' is not a valid rule "\
                          "('on_%s_users' on '%s')",
                          rule, what, obj.absolute_url(1))
                continue
            match_expr = match_expr.strip()
            users_expr = users_expr.strip()
            try:
                if not self._match(match_expr, ec):
                    continue
            except ConflictError:
                raise
            except:
                LOG.error("Error in 'on_%s_users' rule "\
                          "('%s') for '%s'",
                          what, match_expr, obj.absolute_url(1),
                          exc_info=True)
                continue
            if users_expr == '*':
                users.extend(self.getAllUsers())
                break
            else:
                try:
                    users.extend(Expression(users_expr)(ec))
                except ConflictError:
                    raise
                except:
                    LOG.error("Error in 'on_%s_users' rule "\
                              "('%s') for '%s'",
                              what, users_expr, obj.absolute_url(1),
                              exc_info=True)
        return users


    decPrivate('getMailTemplate')
    def getMailTemplate(self, obj, what, ec_bindings=None):
        """Return the template to notify for the ``what`` of an object
        ``obj``, ``what`` being one of the implemented notification
        ("*item_modification*", "*wf_transition*", etc.), or ``None``
        if none could be found.

        ``ec_bindings`` is a mapping which is injected into the
        expression context of the expression of the rules.
        """
        rules = self.getProperty('on_%s_mail_template' % what, None)
        if rules is None:
            raise NotImplementedError, \
                'Notification on "%s" is not implemented.'

        ec = getExpressionContext(obj, ec_bindings)
        template = None
        for rule in rules:
            try:
                match_expr, template_expr = rule.split(RULE_DELIMITER)
                match_expr, template_expr = match_expr.strip(), template_expr.strip()
            except ValueError:
                LOG.error("'%s' is not a valid rule "\
                          "('on_%s_mail_template' on '%s')",
                          rule, what, obj.absolute_url(1))
                continue
            match_expr = match_expr.strip()
            template_expr = template_expr.strip()
            try:
                if not self._match(match_expr, ec):
                    continue
            except ConflictError:
                raise
            except:
                LOG.error("Error in 'on_%s_mail_template' rule "\
                          "('%s') for '%s'",
                          what, match_expr, obj.absolute_url(1),
                          exc_info=True)
                continue
            try:
                template = Expression(template_expr)(ec)
            except ConflictError:
                raise
            except:
                LOG.error("Error in 'on_%s_mail_template' rule "\
                          "('%s') for '%s'",
                          what, template_expr, obj.absolute_url(1),
                          exc_info=True)
                continue
            if type(template) == StringType:
                template = getattr(obj, template, None)
            if template is not None:
                break
        return template


    decPrivate('getAllUsers')
    def getAllUsers(self):
        """Return a list of all user ids of the portal.

        **Warning:** this method may be costly if you rely on an
        external (non ZODB) user source. Use it at your own risk.
        """
        mtool = getToolByName(self, 'portal_membership')
        return mtool.listMemberIds()


    decPrivate('removeUnAuthorizedSubscribers')
    def removeUnAuthorizedSubscribers(self, subscribers, obj):
        """Return users from ``subscribers`` who are authorized to view
        ``obj``.
        """
        portal = getToolByName(self, 'portal_url').getPortalObject()
        mtool = getToolByName(self, 'portal_membership')
        filtered_subscribers = []
        for subscriber in subscribers:
            ## We use '_huntUser()' because 'mtool.getMemberById()'
            ## would have provided a wrapped user object, with a
            ## specific context where the user is not allowed to view
            ## 'obj'. (Damien)
            member = mtool._huntUser(str(subscriber), portal)
            if member is not None:
                if member.has_permission('View', obj):
                    filtered_subscribers.append(subscriber)
            elif self._anonymousShouldBeNotified(obj):
                filtered_subscribers.append(subscriber)
        return filtered_subscribers


    decPrivate('getPreviousVersion')
    def getPreviousVersion(self, obj):
        """Return the previous version of the object, or ``None`` if none
        could be found.

        FIXME: various implementations have been tried, without
        luck. See previous revisions of this file in the SVN
        repository. (Damien)

        Update (06/03/2006): Interesting related informations might be
        found at https://dev.plone.org/archetypes/ticket/648
        """
        return None


    decPrivate('getEmailAddresses')
    def getEmailAddresses(self, users):
        """Return email addresses of ``users``.

        For each value in ``users``:

        - if the value is not an e-mail, suppose it is an user id and
        try to get the ``email`` property of this user;

        - remove duplicates;

        - remove bogus e-mail addresses.
        """
        mtool = getToolByName(self, 'portal_membership')
        addresses = {}
        for user in users:
            member = mtool.getMemberById(str(user))
            if member is not None:
                user = member.getProperty('email')
            if EMAIL_REGEXP.match(user):
                addresses[user] = 1
        return addresses.keys()


    decPrivate('sendNotification')
    def sendNotification(self, addresses, message):
        """Send ``message`` to all ``addresses``."""
        mailhosts = self.superValues(MAIL_HOST_META_TYPES)
        if not mailhosts:
            raise MailHostNotFound
        mailhost = mailhosts[0]

        message = self._encodeSubject(message)

        if self.getProperty('debug_log_addresses'):
            LOG.info('About to send notifications to %s', addresses)

        n_messages_sent = 0
        ## FIXME: This may not be very efficient. Or maybe it
        ## is... Would be good to take a look at other products which
        ## provide a similar feature.
        for address in addresses:
            this_message = ('To: %s\n' % address) + message
            try:
                ## FIXME: SecureMailHost, which is shipped with Plone,
                ## has declared 'send' deprecated and says
                ## 'secureSend' should be used instead. It seems to me
                ## that using '_send()' would be easier. (Damien)
                mailhost.send(this_message)
                n_messages_sent += 1
            except ConflictError:
                raise
            except:
                LOG.error('Error while sending '\
                          'notification: \n%s' % this_message,
                          exc_info=True)
                pass
        return n_messages_sent


    def _match(self, expr, ec):
        """Return ``True`` if ``expr`` returns something which can be
        evaluated to ``True`` in the expression context (``ec``) or if
        ``expr`` is "*".
        """
        if expr == '*':
            return True
        expr = Expression(expr)
        return bool(expr(ec))


    def _getPath(self, obj):
        """Return path of ``obj``.

        A slash ('/') is appended to the path if the object is
        folderish. The returned path is relative to the portal object.
        """
        utool = getToolByName(self, 'portal_url')
        path = utool.getRelativeContentURL(obj)
        path = '/' + path
        if not getattr(obj.aq_base, 'isPrincipiaFolderish', False):
            return path
        if path[-1] != '/':
            path += '/'
        return path


    def _getParents(self, path):
        """Get the parents of the item corresponding to ``path`` and return
        their respective path.

        Parents are returned from ``path`` to the portal root object.
        """
        if path == '/':
            return []
        if path[-1] == '/':
            path = path[:-1]
        parent = path[:path.rfind('/') + 1]
        parents = [parent]
        parents.extend(self._getParents(parent))
        return tuple(parents)


    def _getUID(self, obj):
        """Return UID of the object."""
        portal_uidhandler = getToolByName(self, 'portal_uidhandler')
        uid = portal_uidhandler.queryUid(obj, None)
        if uid is None: ## Not yet registered
            uid = portal_uidhandler.register(obj)
        return uid


    def _anonymousShouldBeNotified(self, obj):
        """Return ``True`` iff anonymous users should be notified.

        It returns ``True`` iff anonymous users have the ``View``
        permission on ``obj``.
        """
        return 'Anonymous' in rolesForPermissionOn('View', obj)


    def _encodeSubject(self, message):
        """Return ``message`` with a correctly encoded subject."""
        ptool = getToolByName(self, 'portal_properties')
        encoding = ptool.site_properties.getProperty('default_charset')
        mout = []
        lines = message.split('\n')
        for l in lines:
            if l.startswith('Subject: '):
                subject = l[9:] ## Remove "Subject: ".
                subject = Header(subject, encoding).encode()
                mout.append('Subject: %s' % subject)
                continue
            mout.append(l)
        return '\n'.join(mout)
    #################################################################


    #################################################################
    ## Extra subscriptions settings
    ###############################
    decProtected('View', 'isExtraSubscriptionsEnabled')
    def isExtraSubscriptionsEnabled(self):
        """Return ``True`` iff extra subscriptions are enabled."""
        return self.getProperty('extra_subscriptions_enabled')


    decProtected('View', 'isExtraSubscriptionsRecursive')
    def isExtraSubscriptionsRecursive(self):
        """Return ``True`` iff extra subscriptions are recursive.

        Note that this method does not check whether extra
        subscriptions are enabled or not.
        """
        return self.getProperty('extra_subscriptions_recursive')


    decProtected('View', 'isExtraSubscriptionsForAuthenticatedOnly')
    def isExtraSubscriptionsForAuthenticatedOnly(self):
        """Return ``True iff extra subscriptions are restricted to
        authenticated users.

        Note that this method does not check whether extra
        subscriptions are enabled or not.
        """
        return self.getProperty('extra_subscriptions_for_authenticated_only')
    #################################################################


    #################################################################
    ## Extra subscriptions logic
    ############################
    def _updateSubscriptionMapping(self, obj):
        """Update subscription mapping."""
        uid = self._getUID(obj)
        path = self._getPath(obj)
        known_path = self._uid_to_path.get(uid)
        if known_path != path:
            self._uid_to_path[uid] = path
            if known_path is not None:
                ## We have old informations for this object
                for key, value in self._subscriptions.items():
                    if key.startswith(known_path):
                        new_key = path + key[len(known_path) : ]
                        self._subscriptions[new_key] = value
                        del self._subscriptions[key]


    decPublic('currentUserHasSubscribePermission')
    def currentUserHasSubscribePermissionOn(self, obj):
        """Return ``True`` iff the current user has ``SUBSCRIBE_PERMISSION``
        on ``obj``.
        """
        mtool = getToolByName(self, 'portal_membership')
        return mtool.checkPermission(SUBSCRIBE_PERMISSION, obj)


    decPublic('subscriptionToParentIsAllowed')
    def isSubscriptionToParentAllowed(self, obj):
        """Return ``True`` iff subscription to the parent of ``obj`` (i.e. the
        first folderish item above ``obj``) is allowed.

        This method uses Plone specific scripts.
        """
        if self.isExtraSubscriptionsRecursive() \
                and not obj.is_folderish() \
                and obj.isDefaultPageInFolder():
            try:
                parent = obj.aq_parent
            except ConflictError:
                raise
            except:
                return False
            return self.currentUserHasSubscribePermissionOn(parent)
        return False


    decPublic('showSubscriptionPortlet')
    def showSubscriptionPortlet(self, obj):
        """Return ``True`` iff the subscription portlet should be shown while
        viewing ``obj``.
        """
        mtool = getToolByName(self, 'portal_membership')
        anonymous = mtool.isAnonymousUser()
        return self.isExtraSubscriptionsEnabled() \
            and not (self.isExtraSubscriptionsForAuthenticatedOnly() \
                     and anonymous) \
                     and self.currentUserHasSubscribePermissionOn(obj)


    decPublic('subscribeTo')
    def subscribeTo(self, obj, email=None):
        """Subscribe ``email`` (or the current user if ``email`` is None) to
        ``obj``.
        """
        if not self.isExtraSubscriptionsEnabled():
            raise DisabledFeature

        if email is not None:
            if not self.isExtraSubscriptionsForAuthenticatedOnly():
                raise DisabledFeature
            if not EMAIL_REGEXP.match(email):
                raise InvalidEmailAddress
            allowed_roles = rolesForPermissionOn(SUBSCRIBE_PERMISSION,
                                                 obj)
            if 'Anonymous' not in allowed_roles:
                raise Unauthorized
            ## FIXME: We would like to send an email to ask the user
            ## to confirm its subscription. Since this has not yet
            ## been implemented, we raise an error. (Damien)
            raise NotImplementedError
        elif not self.currentUserHasSubscribePermissionOn(obj):
            raise Unauthorized
        else:
            user = getSecurityManager().getUser().getId()
            self._updateSubscriptionMapping(obj)
            path = self._getPath(obj)
            subscribers = self._subscriptions.get(path, {})
            subscribers[user] = 1
            self._subscriptions[path] = subscribers


    decPublic('unSubscribeFrom')
    def unSubscribeFrom(self, obj, email=None):
        """Unsubscribe ``email`` (or the current user if ``email`` is
        ``None``) from ``obj``.
        """
        if not self.isExtraSubscriptionsEnabled():
            raise DisabledFeature

        if email is not None:
            if not self.isExtraSubscriptionsForAuthenticatedOnly():
                raise DisabledFeature
            if not EMAIL_REGEXP.match(email):
                raise InvalidEmailAddress
            ## FIXME: an anonymous user would like to unsubscribe
            ## his/her address. This has not yet been implemented, so
            ## we raise an exception. (Damien)
            raise NotImplementedError
        else:
            user = getSecurityManager().getUser().getId()
            self._updateSubscriptionMapping(obj)
            path = self._getPath(obj)
            subscribers = self._subscriptions.get(path, {})
            try:
                del subscribers[user]
                self._subscriptions[path] = subscribers
            except KeyError:
                pass ## User was not subscribed.


    decPublic('unSubscribeFromObjectAbove')
    def unSubscribeFromObjectAbove(self, obj, email=None):
        """Find folderish items above ``obj`` and unsubscribe ``email`` (or
        the current user if ``email`` is ``None``) from the first one
        (s)he is subscribed to.

        If ``user`` is subscribed to ``obj``, this method is equivalent to
        ``unSubscribeFrom(obj, user)``.
        """
        if not self.isExtraSubscriptionsEnabled():
            raise DisabledFeature

        if email is not None:
            if not self.isExtraSubscriptionsForAuthenticatedOnly():
                raise DisabledFeature
            if not EMAIL_REGEXP.match(email):
                raise InvalidEmailAddress
            ## FIXME: an anonymous user would like to unsubscribe
            ## his/her address. This has not yet been implemented, so
            ## we raise an exception. (Damien)
            raise NotImplementedError
        else:
            self._updateSubscriptionMapping(obj)

            utool = getToolByName(obj, 'portal_url')
            portal = utool.getPortalObject()
            portal_container = portal.aq_inner.aq_parent
            while obj != portal_container:
                if self.isSubscribedTo(obj, as_if_not_recursive=True):
                    self.unSubscribeFrom(obj)
                    break
                obj = obj.aq_parent


    decPublic('isSubscribedTo')
    def isSubscribedTo(self, obj, email=None,
                       as_if_not_recursive=False):
        """Return ``True`` iff ``email`` (or the current user if ``email`` is
        ``None``) is subscribed to ``obj``.

        If ``as_if_not_recursive`` is ``True``, this method acts as if
        the recursive mode was off.
        """
        if not self.isExtraSubscriptionsEnabled():
            raise DisabledFeature

        if email is None:
            ## Yes, 'email' is actually the id of the current user.
            email = getSecurityManager().getUser().getId()
        self._updateSubscriptionMapping(obj)
        path = self._getPath(obj)
        subscribers = self.getExtraSubscribersOf(path,
                                                 as_if_not_recursive)
        return subscribers.has_key(email)


    decPrivate('getExtraSubscribersOf')
    def getExtraSubscribersOf(self, path, as_if_not_recursive=False):
        """Return users or email addresses which are subscribed to the given
        path.

        This method returns a mapping whose keys are the users or
        email addresses. If ``as_if_not_recursive`` is ``True``, this
        method acts as if the recursive mode was off.
        """
        subscribers = self._subscriptions.get(path, {}).copy()
        if self.isExtraSubscriptionsRecursive() and \
                not as_if_not_recursive:
            if path[-1] == '/':
                path = path[:-1]
            i = path.rfind('/')
            if i != -1:
                parent = path[:i + 1]
                subscribers.update(self.getExtraSubscribersOf(parent))
        return subscribers
    #################################################################



    decProtected(SUBSCRIBE_PERMISSION, '__useless')
    def __useless(self): pass


InitializeClass(NotificationTool)
