"""
Tests for CMFNotification notification

$Id: testNotification.py 49218 2007-09-14 07:14:19Z dbaty $
"""

import os, sys
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from AccessControl import Unauthorized
from Products.CMFCore.utils import getToolByName
from Products.CMFNotification.exceptions import DisabledFeature
from Products.CMFNotification.NotificationTool import ID as NTOOL_ID
from Products.CMFNotification.permissions import SUBSCRIBE_PERMISSION
from Products.CMFNotification.tests.utils import FakeMailHost
from Products.CMFNotification.tests.utils import CMFNotificationTestCase

## FIXME (Damien): for now, these tests do not actually check email
## contents. That is something I do not really want to do, for now. If
## someone is interested in enhancing these tests, please do.

## FIXME: the current tests are a bit verbose and very similar. They
## could be generated dynamically, but I think debugging would be
## much harder.

class TestNotification(CMFNotificationTestCase):
    """
    Make sure that the notification works, which would be just great,
    since this is the very purpose of this product
    """

    def failUnlessSent(self, n):
        """
        Make the test fail unless 'n' mails have been sent
        """
        mh = self.portal.MailHost
        self.failUnless(len(mh.getSentList()) == n,
                        "Expected %s mail(s), but "\
                        "sent %s" % (n, len(mh.getSentList())))


    def afterSetUp(self):
        """
        Called before each tests

        - replace portal mailhost by a fake one which does not send
        mails but keeps them in a list
        - create users and content.
        """
        portal = self.portal
        self.createTestUsersAndContent()

        ## Set up a fake mail host
        if 'MailHost' in portal.objectIds():
            portal._delObject('MailHost')
        portal._setObject('MailHost', FakeMailHost())


    def testOnItemCreation(self):
        """
        Test notification on item creation
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        wtool = getToolByName(portal, 'portal_workflow')
        mh = portal.MailHost
        self.login('manager')

        ## Disable notification
        ntool.manage_changeProperties(item_creation_notification_enabled=False)
        portal.invokeFactory('Document', 'document')
        portal.document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(0)
        portal.manage_delObjects('document')
        mh.clearSentList()

        ## Enable notification but set the notified users list to []
        ntool.manage_changeProperties(item_creation_notification_enabled=True)
        ntool.manage_changeProperties(on_item_creation_users='* :: python: []')
        portal.invokeFactory('Document', 'document')
        portal.document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(0)
        portal.manage_delObjects('document')
        mh.clearSentList()

        ## Set the notified users list to <everybody> but ask for a
        ## missing mail template
        ntool.manage_changeProperties(on_item_creation_users='* :: *')
        ntool.manage_changeProperties(on_item_creation_mail_template='* :: string:does_not_exist')
        portal.invokeFactory('Document', 'document')
        portal.document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(0)
        portal.manage_delObjects('document')
        mh.clearSentList()

        ## Set the default mail template
        ntool.manage_changeProperties(on_item_creation_mail_template='* :: string:at_notification')
        portal.invokeFactory('Document', 'document')
        portal.document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(3)
        portal.manage_delObjects('document')
        mh.clearSentList()

        ## Set 'plone_workflow' initial state to 'private', thus
        ## hiding the ne item from everybody but 'manager'.
        wtool.plone_workflow.initial_state = 'private'
        portal.invokeFactory('Document', 'document')
        portal.document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(1)
        portal.manage_delObjects('document')
        mh.clearSentList()


    def testOnItemModification(self):
        """
        Test notification on item modification
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        wtool = getToolByName(portal, 'portal_workflow')
        document = portal.folder.document1
        mh = portal.MailHost
        self.login('manager')

        ## Disable notification
        ntool.manage_changeProperties(item_modification_notification_enabled=False)
        document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Enable notification but set the notified users list to []
        ntool.manage_changeProperties(item_modification_notification_enabled=True)
        ntool.manage_changeProperties(on_item_modification_users='* :: python: []')
        document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Set the notified users list to <everybody> but ask for a
        ## missing mail template
        ntool.manage_changeProperties(on_item_modification_users='* :: *')
        ntool.manage_changeProperties(on_item_modification_mail_template='* :: string:does_not_exist')
        document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Set the default mail template
        ntool.manage_changeProperties(on_item_modification_mail_template='* :: string:at_notification')
        document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(3)
        mh.clearSentList()

        ## Hide 'document' from everybody except manager
        wtool.doActionFor(document, 'hide')
        document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(1)
        mh.clearSentList()


    def testOnWorkflowTransition(self):
        """
        Test notification on workflow transition
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        wtool = getToolByName(portal, 'portal_workflow')
        document = portal.folder.document1
        mh = portal.MailHost
        self.login('manager')

        ## Disable notification
        ntool.manage_changeProperties(wf_transition_notification_enabled=False)
        wtool.doActionFor(document, 'publish')
        self.failUnlessSent(0)
        wtool.doActionFor(document, 'retract')
        mh.clearSentList()

        ## Enable notification but set the notified users list to []
        ntool.manage_changeProperties(wf_transition_notification_enabled=True)
        ntool.manage_changeProperties(on_wf_transition_users='* :: python: []')
        wtool.doActionFor(document, 'publish')
        self.failUnlessSent(0)
        wtool.doActionFor(document, 'retract')
        mh.clearSentList()

        ## Set the notified users list to <everybody> but ask for a
        ## missing mail template
        ntool.manage_changeProperties(on_wf_transition_users='* :: *')
        ntool.manage_changeProperties(on_wf_transition_mail_template='* :: string:does_not_exist')
        wtool.doActionFor(document, 'publish')
        self.failUnlessSent(0)
        wtool.doActionFor(document, 'retract')
        mh.clearSentList()

        ## Set the default mail template
        ntool.manage_changeProperties(on_wf_transition_mail_template='* :: string:at_notification')
        wtool.doActionFor(document, 'publish')
        self.failUnlessSent(3)
        wtool.doActionFor(document, 'retract')
        mh.clearSentList()

        ## Hide 'document' from everybody except manager
        wtool.doActionFor(document, 'hide')
        self.failUnlessSent(1)
        mh.clearSentList()


    def testOnMemberRegistration(self):
        """
        Test notification on member registration
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        rtool = getToolByName(portal, 'portal_registration')
        mtool = getToolByName(portal, 'portal_membership')
        deleteMember = lambda userid: mtool.deleteMembers([userid, ])
        mh = portal.MailHost
        userid = 'a_new_member'
        self.login('manager')

        ## Disable notification
        ntool.manage_changeProperties(member_registration_notification_enabled=False)
        rtool.addMember(userid, 'password', properties=None)
        self.failUnlessSent(0)
        deleteMember(userid)
        mh.clearSentList()

        ## Enable notification but set the notified users list to []
        ntool.manage_changeProperties(member_registration_notification_enabled=True)
        ntool.manage_changeProperties(on_member_registration_users='* :: python: []')
        rtool.addMember(userid, 'password', properties=None)
        self.failUnlessSent(0)
        deleteMember(userid)
        mh.clearSentList()

        ## Set the notified users list to <everybody> but ask for a
        ## missing mail template
        ntool.manage_changeProperties(on_member_registration_users='* :: *')
        ntool.manage_changeProperties(on_member_registration_mail_template='* :: string:does_not_exist')
        rtool.addMember(userid, 'password', properties=None)
        self.failUnlessSent(0)
        deleteMember(userid)
        mh.clearSentList()

        ## Set the default mail template
        ntool.manage_changeProperties(on_member_registration_mail_template='* :: string:member_notification')
        rtool.addMember(userid, 'password', properties=None)
        self.failUnlessSent(3)
        deleteMember(userid)
        mh.clearSentList()


    def testOnMemberModification(self):
        """
        Test notification on member modification
        """
        pass ## FIXME: implement me! (Damien)


    def testOnDiscussionItemCreation(self):
        """
        Test notification on discussion item creation
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        wtool = getToolByName(portal, 'portal_workflow')
        document = portal.folder.document1
        document.allow_discussion = True
        mh = portal.MailHost
        addDiscussionItem = lambda obj: obj.discussion_reply('Subject',
                                                             'Body text')
        self.login('manager')

        ## Disable notification
        ntool.manage_changeProperties(on_discussion_item_creation_notification_enabled=False)
        addDiscussionItem(document)
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Enable notification but set the notified users list to []
        ntool.manage_changeProperties(discussion_item_creation_notification_enabled=True)
        ntool.manage_changeProperties(on_discussion_item_creation_users='* :: python: []')
        addDiscussionItem(document)
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Set the notified users list to <everybody> but ask for a
        ## missing mail template
        ntool.manage_changeProperties(on_discussion_item_creation_users='* :: *')
        ntool.manage_changeProperties(on_discussion_item_creation_mail_template='* :: string:does_not_exist')
        addDiscussionItem(document)
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Set the default mail template
        ntool.manage_changeProperties(on_discussion_item_creation_mail_template='* :: string:discussion_notification')
        addDiscussionItem(document)
        self.failUnlessSent(3)
        mh.clearSentList()

        ## Hide 'document' from everybody except manager
        wtool.doActionFor(document, 'hide')
        addDiscussionItem(document)
        self.failUnlessSent(1)
        mh.clearSentList()

        ## Reply to a discussion item
        discussion_item = document.talkback.objectValues()[0]
        discussion_item = discussion_item.__of__(document.talkback)
        addDiscussionItem(discussion_item)
        self.failUnlessSent(1)
        mh.clearSentList()


    def testHandlersDoNotOverlap(self):
        """
        Test that handlers do not overlap themselves, i.e. that at
        most one handler is called for any event
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        wtool = getToolByName(portal, 'portal_workflow')
        rtool = getToolByName(portal, 'portal_registration')
        mtool = getToolByName(portal, 'portal_membership')
        deleteMember = lambda userid: mtool.deleteMembers([userid, ])
        mh = portal.MailHost
        document = portal.folder.document1
        userid = 'a_new_member'

        self.login('manager')

        ## Enable everything
        users = '* :: *'
        item_template = '* :: string:at_notification'
        member_template = '* :: string:member_notification'
        discussion_template = '* :: string:discussion_notification'
        properties = {
           'item_creation_notification_enabled': True,
           'on_item_creation_users': users,
           'on_item_creation_mail_template': item_template,
           'item_modification_notification_enabled': True,
           'on_item_modification_users': users,
           'on_item_modification_mail_template': item_template,
           'wf_transition_notification_enabled': True,
           'on_wf_transition_users': users,
           'on_wf_transition_mail_template': item_template,
           'member_registration_notification_enabled': True,
           'on_member_registration_users': users,
           'on_member_registration_mail_template': member_template,
           'discussion_item_creation_notification_enabled': True,
           'on_discussion_item_creation_users': users,
           'on_discussion_item_creation_mail_template': discussion_template,
           }
        ntool.manage_changeProperties(properties)

        ## Test item creation
        portal.invokeFactory('Document', 'document')
        portal.document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(3)
        portal.manage_delObjects('document')
        mh.clearSentList()

        ## Test item modification
        document.processForm(data=1,
                             values={'title': 'New title'})
        self.failUnlessSent(3)
        mh.clearSentList()

        ## Test workflow transition
        wtool.doActionFor(document, 'publish')
        self.failUnlessSent(3)
        wtool.doActionFor(document, 'retract')
        mh.clearSentList()

        ## Test member registration
        rtool.addMember(userid, 'password', properties=None)
        self.failUnlessSent(3)
        deleteMember(userid)
        mh.clearSentList()

        ## Test member modification (FIXME: Damien)

        ## Test discussion item creation
        document.allow_discussion = True
        document.discussion_reply('Subject', 'Body text')
        self.failUnlessSent(3)
        mh.clearSentList()


if  __name__ == '__main__':
    framework()

def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestNotification))
    return suite
